import bz2
import os
from datetime import datetime
import _pickle

DC_VERSION = '3.4.0'


class FileBase:
    def __init__(self,
                 subject_id=None, protocol_name=None, task_name=None, fatigue=None, hungriness=None, sequence=None,
                 project=None, time=None, description=None, gain=24, tags=None, leadoff=None, line=None):
        if leadoff is None:
            leadoff = []
        if tags is None:
            tags = []
        self.line = line
        self.leadoff = leadoff
        self.tags = tags
        self.gain = gain
        self.subject = subject_id
        self.protocol = protocol_name
        self.task = task_name
        self.description = description
        self.fatigue = fatigue
        self.hungriness = hungriness
        self.sequence = sequence
        self.project = project
        self.timestamp = time if time else datetime.now()
        self.filename = str(int(datetime.timestamp(self.timestamp))) + '_' + str(self.task) + '_' + str(self.subject)
        self.device_data = None
        self.brain_data = None
        self.extra_signal = None
        self.output_data = None
        self.reaction_data = None

    # SETTERS

    def set_leadoff(self, leadoff):
        self.leadoff = leadoff

    def set_line(self, line):
        self.line = line

    def set_tags(self, tags):
        self.tags = tags

    def set_gain(self, gain):
        self.gain = gain

    def set_subject(self, subject_id):
        self.subject = subject_id

    def set_protocol(self, protocol_name):
        self.protocol = protocol_name

    def set_description(self, description):
        self.description = description

    def set_sequence(self, sequence):
        self.sequence = sequence

    def set_project(self, project):
        self.project = project

    def set_hungriness(self, hungriness):
        self.hungriness = hungriness

    def set_fatigue(self, fatigue):
        self.fatigue = fatigue

    def set_task(self, task_name):
        self.task = task_name

    def set_timestamp(self, timestamp):
        self.timestamp = timestamp

    def set_reaction_data(self, feedback, label, correct_answers=None):
        if self.brain_data:
            self.reaction_data = ReactionData(feedback, label)
            return self.reaction_data
        else:
            return None

    def set_feedback(self, feedback):
        self.reaction_data.set_feedback(feedback)

    def set_label(self, label):
        self.reaction_data.set_label(label)

    def set_device_data(self, device_name, software_version, firmware_version, sensor_name):
        self.device_data = DeviceData(device_name, software_version, firmware_version, sensor_name)
        return self.device_data

    def set_sensor(self, sensor_name):
        self.device_data.set_sensor(sensor_name)
        return self.device_data

    def set_brain_data(self, raw_data, channel_count, sample_count, sample_rate, channels_name=None):
        self.brain_data = BrainData(raw_data, channel_count, sample_count, sample_rate, channels_name)
        return self.brain_data

    def set_brain_channels_name(self, names):
        if self.brain_data:
            return self.brain_data.set_channels_name(names)
        else:
            return False

    def set_brain_channel_name(self, channel, name):
        if self.brain_data:
            return self.brain_data.set_channel_name(channel, name)
        else:
            return False

    def set_extra_signal(self, raw_data, channel_count, sample_count, sample_rate, channels_name=None):
        self.extra_signal = ExtraSignal(raw_data, channel_count, sample_count, sample_rate, channels_name)
        return self.extra_signal

    def set_extra_channels_name(self, names):
        if self.extra_signal:
            return self.extra_signal.set_channels_name(names)
        else:
            return False

    def set_extra_channel_name(self, channel, name):
        if self.extra_signal:
            return self.extra_signal.set_channel_name(channel, name)
        else:
            return False

    def set_output_data(self, raw_data, channel_count, sample_count, sample_rate):
        self.output_data = OutputData(raw_data, channel_count, sample_count, sample_rate)
        return self.output_data

    # FILE WORLD

    def load(self, filename, compress=False):
        if os.path.isfile(filename):
            try:
                if compress:
                    f = bz2.BZ2File(filename, 'rb')
                else:
                    f = open(filename, 'rb')
                tmp_dict = _pickle.load(f)
                f.close()
                self.__dict__.update(tmp_dict)
                return 'OK'
            except Exception as e:
                print(e)
                return 'File can not load.'
        return 'File not exist.'

    def load_from_file(self, file, compress=False):
        try:
            if compress:
                file = bz2.BZ2File(file)
            tmp_dict = _pickle.load(file)
            self.__dict__.update(tmp_dict)
            return 'OK'
        except Exception as e:
            print(e)
            return 'File can not open.'

    def save(self, path_to_file, filename=None, compress=False):
        if not os.path.exists(path_to_file):
            os.mkdir(path_to_file)

        if filename is None:
            filename = self.filename
        file = str(path_to_file + filename) + ".iec"

        try:
            if compress:
                with bz2.BZ2File(file, 'w') as f:
                    _pickle.dump(self.__dict__, f, 4)
            else:
                with open(file, 'wb+') as f:
                    _pickle.dump(self.__dict__, f, 4)

            return self.get_filename()
        except Exception as e:
            print(e)
            return 'File can not save.'

    # GETTERS

    def get_leadoff(self):
        return self.leadoff

    def get_line(self):
        return self.line

    def get_tags(self):
        return self.tags

    def get_gain(self):
        return self.gain

    def get_subject_id(self):
        return self.subject

    def get_protocol_name(self):
        return self.protocol

    def get_task_name(self):
        return self.task

    def get_description(self):
        return self.description

    def get_sequence(self):
        return self.sequence

    def get_project(self):
        return self.project

    def get_hungriness(self):
        return self.hungriness

    def get_fatigue(self):
        return self.fatigue

    def get_timestamp(self):
        return self.timestamp

    def get_filename(self):
        return self.filename

    # output
    def get_output_raw_data(self):
        if self.output_data:
            return self.output_data.get_raw_data()
        else:
            return None

    def get_output_channel_count(self):
        if self.output_data:
            return self.output_data.get_channel_count()
        else:
            return None

    def get_output_sample_count(self):
        if self.output_data:
            return self.output_data.get_sample_count()
        else:
            return None

    def get_output_sample_rate(self):
        if self.output_data:
            return self.output_data.get_sample_rate()
        else:
            return None

    # extra signal
    def get_extra_raw_data(self):
        if self.extra_signal:
            return self.extra_signal.get_raw_data()
        else:
            return None

    def get_extra_channel_count(self):
        if self.extra_signal:
            return self.extra_signal.get_channel_count()
        else:
            return None

    def get_extra_sample_count(self):
        if self.extra_signal:
            return self.extra_signal.get_sample_count()
        else:
            return None

    def get_extra_sample_rate(self):
        if self.extra_signal:
            return self.extra_signal.get_sample_rate()
        else:
            return None

    def get_extra_channels_name(self):
        if self.extra_signal:
            return self.extra_signal.get_channels_name()
        else:
            return None

    def get_extra_channel_name(self, channel):
        if self.extra_signal:
            return self.extra_signal.get_channel_name(channel)
        else:
            return None

    # brain data
    def get_brain_raw_data(self):
        if self.brain_data:
            return self.brain_data.get_raw_data()
        else:
            return None

    def get_brain_channel_count(self):
        if self.brain_data:
            return self.brain_data.get_channel_count()
        else:
            return None

    def get_brain_sample_count(self):
        if self.brain_data:
            return self.brain_data.get_sample_count()
        else:
            return None

    def get_brain_sample_rate(self):
        if self.brain_data:
            return self.brain_data.get_sample_rate()
        else:
            return None

    def get_brain_channels_name(self):
        if self.brain_data:
            return self.brain_data.get_channels_name()
        else:
            return None

    def get_brain_channel_name(self, channel):
        if self.brain_data:
            return self.brain_data.get_channel_name(channel)
        else:
            return None

    # device and sensor
    def get_device_info(self):
        if self.device_data:
            return self.device_data.get_device_info()
        else:
            return None

    def get_sensor_name(self):
        if self.device_data:
            return self.device_data.get_sensor_name()
        else:
            return None

    # reaction data
    def get_feedback_data(self):
        if self.reaction_data:
            return self.reaction_data.get_feedback()
        else:
            return None

    def get_label_data(self):
        if self.reaction_data:
            return self.reaction_data.get_label()
        else:
            return None


class OutputData:
    def __init__(self, raw_data, channel_count, sample_count, sample_rate):
        self.raw_data = raw_data
        self.channel_count = channel_count
        self.sample_count = sample_count
        self.sample_rate = sample_rate

    def get_raw_data(self):
        return self.raw_data

    def get_channel_count(self):
        return self.channel_count

    def get_sample_count(self):
        return self.sample_count

    def get_sample_rate(self):
        return self.sample_rate


class ExtraSignal:
    def __init__(self, raw_data, channel_count=0, sample_count=0, sample_rate=0, channels_name=None):
        if channels_name is None:
            channels_name = []
            for i in range(channel_count):
                channels_name.append('')

        self.channels_name = channels_name
        self.raw_data = raw_data
        self.channel_count = channel_count
        self.sample_count = sample_count
        self.sample_rate = sample_rate

    def get_raw_data(self):
        return self.raw_data

    def get_channel_count(self):
        return self.channel_count

    def get_channels_name(self):
        return self.channels_name

    def get_channel_name(self, channel):
        if channel < len(self.channels_name):
            return self.channels_name[channel]
        else:
            return None

    def set_channels_name(self, names):
        if isinstance(names, list) and len(names) == self.channel_count:
            self.channels_name = names
            return True
        else:
            return False

    def set_channel_name(self, channel, name):
        if not self.channels_name:
            channels_name = []
            for i in range(self.channel_count):
                channels_name.append('')

        if channel < self.channel_count:
            self.channels_name[channel] = name
            return True
        else:
            return False

    def get_sample_count(self):
        return self.sample_count

    def get_sample_rate(self):
        return self.sample_rate


class BrainData:
    def __init__(self, raw_data, channel_count=0, sample_count=0, sample_rate=0, channels_name=None):
        if channels_name is None:
            channels_name = []
            for i in range(channel_count):
                channels_name.append('')

        self.channels_name = channels_name
        self.raw_data = raw_data
        self.channel_count = channel_count
        self.sample_count = sample_count
        self.sample_rate = sample_rate

    def get_raw_data(self):
        return self.raw_data

    def get_channel_count(self):
        return self.channel_count

    def get_channels_name(self):
        return self.channels_name

    def get_channel_name(self, channel):
        if channel < len(self.channels_name):
            return self.channels_name[channel]
        else:
            return None

    def set_channels_name(self, names):
        if isinstance(names, list) and len(names) == self.channel_count:
            self.channels_name = names
            return True
        else:
            return False

    def set_channel_name(self, channel, name):
        if not self.channels_name:
            channels_name = []
            for i in range(self.channel_count):
                channels_name.append('')

        if channel < self.channel_count:
            self.channels_name[channel] = name
            return True
        else:
            return False

    def get_sample_count(self):
        return self.sample_count

    def get_sample_rate(self):
        return self.sample_rate


class DeviceData:
    def __init__(self, device_name, software_version, firmware_version, sensor_name):
        self.device_name = device_name
        self.software_version = software_version
        self.firmware_version = firmware_version
        self.sensor_name = sensor_name

    def get_device_info(self):
        return {
            'device_name': self.device_name,
            'software_version': self.software_version,
            'firmware_version': self.firmware_version
        }

    def get_sensor_name(self):
        return self.sensor_name

    def set_sensor(self, sensor_name):
        self.sensor_name = sensor_name


class ReactionData:
    def __init__(self, feedback, label):
        self.feedback = feedback
        self.label = label

    def get_feedback(self):
        return self.feedback

    def get_label(self):
        return self.label

    def set_feedback(self, feedback):
        self.feedback = feedback

    def set_label(self, label):
        self.label = label
